/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.framework;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

/**
 * Component configuration. The JBI extensions are saved to a properties file
 * during the bootstrap phase. These properties are then used by the component
 * for configuration purpose.
 *
 * @author msauvage chamerling - eBMWebSourcing
 *
 */
public class ComponentPropertiesManager {

    public static final String POOL_SIZE = "pool-size";

    public static final String DEFAULT_POOL_SIZE = "10";

    public static final String IGNORED_STATUS = "ignored-status";

    public static final String DEFAULT_IGNORED_STATUS = "DONE_AND_ERROR_IGNORED";

    public static final String JBI_LISTENER_CLASSNAME = "jbi-listener-class-name";

    public static final String EXTERNAL_LISTENER_CLASSNAME = "external-listener-class-name";

    public static final String MANAGED_METHOD_PREFIX = "managedMethod";

    public static final String ACKNOWLEDGEMENT_TIMEOUT = "acknowledgement-timeout";

    public static final String DEFAULT_ACKNOWLEDGEMENT_TIMEOUT = "5000";

    /**
     * The default properties file name
     */
    public static final String PROPERTIES_FILE = "component.properties";

    /**
     * The storage object
     */
    protected Properties properties;

    /**
     * The root path of the properties file. It is also the root path of the
     * component installation.
     */
    protected String rootPath;

    /**
     * Creates a new instance of {@link ComponentPropertiesManager}
     *
     */
    public ComponentPropertiesManager() {
        this.properties = new Properties();
    }

    /**
     *
     *
     */
    public void clear() {
        this.properties = new Properties();
    }

    /**
     * Get the property value or return the default value if the property has
     * not been found.
     *
     * @param key
     * @param defaultValue
     * @return
     */
    public String getProperty(String key, String defaultValue) {
        return this.properties.getProperty(key, defaultValue);
    }

    /**
     * Returns the value corresponding to the key in the properties as a String.
     *
     * @param key
     *            the key searched
     * @return a String representation of the value corresponding to the key,
     *         null if key doesn't exists
     */
    public String getProperty(String key) {
        return properties.getProperty(key);
    }

    /**
     * Get the exposed attributes names
     *
     * @return
     */
    public Set<String> getExposedAttributeKeys() {
        SortedSet<String> set = new TreeSet<String>();
        for (Object name : getProperties().keySet()) {
            String key = (String) name;
            if (!key
                    .startsWith(ComponentPropertiesManager.MANAGED_METHOD_PREFIX)) {
                set.add(key);
            }
        }
        return set;
    }

    /**
     * Get the exposed method names
     *
     * @return
     */
    public List<String> getExposedMethodNames() {
        List<String> list = new ArrayList<String>();
        for (Object name : getProperties().keySet()) {
            String key = (String) name;
            if (key
                    .startsWith(ComponentPropertiesManager.MANAGED_METHOD_PREFIX)) {
                list.add(getProperty(key));
            }
        }
        return list;
    }

    /**
     * Returns all the properties
     *
     * @return the properties as a Properties object
     */
    public Properties getProperties() {
        return this.properties;
    }

    /**
     * Load the default properties for all the components.
     *
     */
    protected final void initDefaultValues() {
        // thread pool properties
        if (getProperty(POOL_SIZE) == null) {
            setAttribute(POOL_SIZE, DEFAULT_POOL_SIZE);
        }

        if (getProperty(IGNORED_STATUS) == null) {
            setAttribute(IGNORED_STATUS, DEFAULT_IGNORED_STATUS);
        }

        if (getProperty(ACKNOWLEDGEMENT_TIMEOUT) == null) {
            setAttribute(ACKNOWLEDGEMENT_TIMEOUT, DEFAULT_ACKNOWLEDGEMENT_TIMEOUT);
        }

        // additional default properties
        // TBD...
    }

    /**
     * Load properties from file. If file is not found, try to get one from the
     * classpath.
     *
     * @param file
     *            the file to read
     */
    public boolean load() {
        File f = null;
        InputStream is = null;
        boolean result = true;

        if (rootPath != null) {

            // try to get component properties from the component root path
            f = new File(rootPath, PROPERTIES_FILE);
            if (!f.exists()) {
                f = null;
            }

            // try to get component properties from classpath
            if (f == null) {
                is = this.getClass().getClassLoader().getResourceAsStream(
                        PROPERTIES_FILE);
            }

            try {
                // load properties
                if (f != null) {
                    properties.load(new FileInputStream(f));
                } else if (is != null) {
                    properties.load(is);
                } else {
                    result = false;
                }
            } catch (FileNotFoundException e) {
                // Nothing to do
            } catch (IOException e) {
                // Nothing to do
            }
        }
        return result;
    }

    /**
     *
     * @return
     */
    public Properties retrievePersistedProperties() {
        File f = null;
        Properties properties = new Properties();

        if (rootPath != null) {
            // try to get component properties from the component root path
            f = new File(rootPath, PROPERTIES_FILE);
            if (!f.exists()) {
                f = null;
            }

            try {
                // load properties
                if (f != null) {
                    properties.load(new FileInputStream(f));
                }
            } catch (FileNotFoundException e) {
                // Nothing to do
            } catch (IOException e) {
                // Nothing to do
            }
        }
        return properties;
    }

    /**
     * Save the actual properties into a file
     *
     * @param file
     *            the complete path to the file to write
     */
    public void save() {
        FileOutputStream os = null;
        try {
            os = new FileOutputStream(new File(rootPath, PROPERTIES_FILE));
            properties.store(os, "Component extensions");
        } catch (IOException e) {
            // Nothing to do
        } finally {
            if (os != null) {
                try {
                    os.close();
                } catch (IOException e) {
                    // Nothing to do
                }
            }
        }
    }

    /**
     * Sets a new key/value property. If property already exist, value is
     * updated
     *
     * @param key
     *            the key of the property
     * @param value
     *            the value corresponding
     */
    public void setAttribute(String key, String value) {
        if (value == null) {
            properties.remove(key);
        } else {
            properties.setProperty(key, value);
        }
    }

    /**
     * Copies all of the mappings from the specified Map to the properties
     * attribute.
     *
     * @param properties
     *            a Map of the <key,value> pairs to be stored
     */
    public void setProperties(Properties properties) {
        this.properties = properties;
    }

    /**
     *
     * @param methods
     */
    public void setExposedMethods(List<String> methods) {
        if (methods != null) {
            for (String method : methods) {
                properties.setProperty(MANAGED_METHOD_PREFIX + "." + method,
                        method);
            }
        }
    }

    /**
     *
     * @param rootPath
     */
    public void setRootPath(String rootPath) {
        this.rootPath = rootPath;
    }
}
