/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005-2006 EBM WebSourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: SourceHelper.java 94 2006-03-26 17:11:05Z alouis $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.framework.util;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.URI;
import java.util.Properties;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.objectweb.petals.component.framework.PEtALSCFException;
import org.objectweb.petals.component.framework.exception.DocumentException;
import org.objectweb.petals.component.framework.exception.FaultException;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Sample class used to manage the mapping between
 * <code>javax.xml.transform.Source</code> from String. the mapping is a
 * quicly "string" -> "<content>string</content>" conversion. Initial
 * developer(s): Adrien LOUIS
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 */
public final class SourceHelper {

	private static final String START_TAG = "<content>";

	private static final String END_TAG = "</content>";

	private static final int SOAP_FAULT_BUFFER_SIZE = 2048;

	private final static ThreadLocal<DocumentBuilder> documentBuilderThreadLocal = new ThreadLocal<DocumentBuilder>() {

        @Override
        protected DocumentBuilder initialValue() {
            try {
                final DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory
                        .newInstance();
                documentBuilderFactory.setNamespaceAware(true);
                final DocumentBuilder documentBuilder = documentBuilderFactory
                        .newDocumentBuilder();
                return documentBuilder;
            } catch (ParserConfigurationException e) {
                throw new RuntimeException("Failed to create DocumentBuilder",
                        e);
            }
        }
    };
    
    private final static ThreadLocal<Transformer> transformerWithoutXmlDeclarationThreadLocal = new ThreadLocal<Transformer>() {

        @Override
        protected Transformer initialValue() {
            try {

                final Transformer transformer = TransformerFactory.newInstance().newTransformer();
                Properties props = new Properties();
                props.put(OutputKeys.OMIT_XML_DECLARATION, "yes");
                transformer.setOutputProperties(props);
                return transformer;

            } catch (TransformerConfigurationException e) {
                throw new RuntimeException("Failed to create Transformer",
                        e);
            }
        }
    };

	/**
	 * Creates a new instance of {@link SourceHelper}
	 * 
	 */
	private SourceHelper() {
		super();
	}

	/**
	 * Create a Source from String adding start and end tags.
	 * 
	 * @param msg
	 * @return
	 * @throws PEtALSCFException
	 * @deprecated Don't use this method. Prefer ${@link #createSource(Document)}
	 */
	@Deprecated
    public static Source createContentSource(final String msg)
			throws PEtALSCFException {
		return createSource(START_TAG + msg + END_TAG);
	}
	
    /**
     * Creates an XML fault with the stacktrace, message and the specified code
     * 
     * @param e
     *            exception, used to provide faultstring and detail
     * @param code
     *            code of the fault
     * @return A string containing the xml code that corresponds to the Soap
     *         Fault
     * @deprecated use method createSoapFaultDocument(final Throwable e, final
     *             QName qname) instead
     */
	@Deprecated
    public static String createSoapFault(final Throwable e, final String code) {
        return SourceHelper.createSoapFault(e, new QName(code));
    }

    /**
     * <p>
     * Creates an SOAP fault according to provided exception and code:<br>
     * <code>&lt;s:Fault xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;faultcode&gt;{org.objectweb.petals}ERROR&lt;/faultcode&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;faultstring&gt;mesage&lt;/faultstring&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;faultactor&gt;component&lt;/faultactor&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;detail&gt;<br>
     * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;...<br>
     * &nbsp;&nbsp;&nbsp;&lt;/detail&gt;<br>
     * &lt;s:Fault&gt;</code>
     * </p>
     * 
     * <p>
     * If the provided exception is an instance of {@link DocumentException}:
     * <ul>
     * <li>the fault code is set with the value returned by
     * {@link DocumentException#getFaultCode()},</li>
     * <li>the fault string is set with the value returned by
     * {@link DocumentException#getMessage()},</li>
     * <li>the fault actor is set with the value returned by
     * {@link DocumentException#getFaultActor()},</li>
     * <li>the detail is set with the exception document returned by
     * {@link DocumentException#getDocument()()}.</li>
     * </ul>
     * </p>
     * 
     * <p>
     * If the provided exception is an instance of {@link FaultException}:
     * <ul>
     * <li>the fault code is set with the value returned by
     * {@link FaultException#getFaultCode()},</li>
     * <li>the fault string is set with the value returned by
     * {@link FaultException#getMessage()},</li>
     * <li>the fault actor is set with the value returned by
     * {@link FaultException#getFaultActor()},</li>
     * <li>the detail is set with the exception stack trace.</li>
     * </ul>
     * </p>
     * 
     * <p>
     * otherwise:
     * <ul>
     * <li>the fault code is set with the value of <code>errorCode</code>,</li>
     * <li>the fault string is set with the value returned by
     * {@link Exception#getMessage()},</li>
     * <li>the fault actor is set to <code>null</code>,</li>
     * <li>the detail is set with the exception stack trace.</li>
     * </ul>
     * </p>
     * 
     * @param e
     *            An exception used to provide faultstring and detail
     * @param errorCode
     *            The QName of the fault code. It is not used if the exception
     *            is an instance of {@link FaultException} or
     *            {@link DocumentException}.
     * @return A string containing the xml code that corresponds to the Soap
     *         Fault, using the encoding UTF-8.
     * @deprecated use the method ${link
     *             {@link #createSoapFaultDocument(Throwable, QName)} instead.
     */
    @Deprecated
    public static String createSoapFault(final Throwable e,
            final QName errorCode) {
        try {
            Document faultDocument = createSoapFaultDocument(e, errorCode);
            
            return XMLUtil.createStringFromDOMDocument(faultDocument);
            
        } catch (Exception e1) {
            StringBuffer s = new StringBuffer(SourceHelper.SOAP_FAULT_BUFFER_SIZE);
            s.append("<s:Fault xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\">");
            s.append("<faultcode>s:Server</faultcode>");
            s.append("<faultstring>").append(e1.getMessage()).append("</faultstring>");
            s.append("</s:Fault>");
            return s.toString();
        }
        
    }

    /**
     * <p>
     * Creates an SOAP fault, as a DOM document, according to provided exception
     * and code:<br>
     * <code>&lt;s:Fault xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;faultcode&gt;{org.objectweb.petals}ERROR&lt;/faultcode&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;faultstring&gt;mesage&lt;/faultstring&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;faultactor&gt;component&lt;/faultactor&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;detail&gt;<br>
     * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;...<br>
     * &nbsp;&nbsp;&nbsp;&lt;/detail&gt;<br>
     * &lt;s:Fault&gt;</code>
     * </p>
     * 
     * <p>
     * If the provided exception is an instance of {@link DocumentException}:
     * <ul>
     * <li>the fault code is set with the value returned by
     * {@link DocumentException#getFaultCode()},</li>
     * <li>the fault string is set with the value returned by
     * {@link DocumentException#getMessage()},</li>
     * <li>the fault actor is set with the value returned by
     * {@link DocumentException#getFaultActor()},</li>
     * <li>the detail is set with the exception document returned by
     * {@link DocumentException#getDocument()()}.</li>
     * </ul>
     * </p>
     * 
     * <p>
     * If the provided exception is an instance of {@link FaultException}:
     * <ul>
     * <li>the fault code is set with the value returned by
     * {@link FaultException#getFaultCode()},</li>
     * <li>the fault string is set with the value returned by
     * {@link FaultException#getMessage()},</li>
     * <li>the fault actor is set with the value returned by
     * {@link FaultException#getFaultActor()},</li>
     * <li>the detail is set with the exception stack trace.</li>
     * </ul>
     * </p>
     * 
     * <p>
     * otherwise:
     * <ul>
     * <li>the fault code is set with the value of <code>errorCode</code>,</li>
     * <li>the fault string is set with the value returned by
     * {@link Exception#getMessage()},</li>
     * <li>the fault actor is set to <code>null</code>,</li>
     * <li>the detail is set with the exception stack trace.</li>
     * </ul>
     * </p>
     * 
     * @param e
     *            An exception used to provide faultstring and detail
     * @param errorCode
     *            The QName of the fault code. It is not used if the exception
     *            is an instance of {@link FaultException} or
     *            {@link DocumentException}.
     * @return A string containing the xml code that corresponds to the Soap
     *         Fault
     */
    public static Document createSoapFaultDocument(final Throwable e,
            final QName errorCode) {

        QName faultCode = errorCode;
        String faultString = e.getMessage();
        String details = "";
        URI faultActor = null;
        if (e instanceof DocumentException) {
            faultCode = ((DocumentException) e).getFaultCode();
            faultActor = ((DocumentException) e).getFaultActor();
            final Document document = ((DocumentException) e).getDocument();
            if ((document != null) && (document.getDocumentElement() != null)) {
                try {
                    details = XMLUtil.parseToString(document
                            .getDocumentElement());
                } catch (TransformerException e1) {
                    details = "An error occured when retrieving the exception document ("
                            + e1.getMessage() + ").";
                }
            }
        } else if (e instanceof FaultException) {
            faultCode = ((FaultException) e).getFaultCode();
            faultActor = ((FaultException) e).getFaultActor();
            
            final StringWriter sw = new StringWriter();
            sw.append("<![CDATA[");
            final PrintWriter pw = new PrintWriter(sw);
            e.printStackTrace(pw);
            sw.append("]]>");

            details = sw.toString();

        } else {
            final StringWriter sw = new StringWriter();
            sw.append("<![CDATA[");
            final PrintWriter pw = new PrintWriter(sw);
            e.printStackTrace(pw);
            sw.append("]]>");

            details = sw.toString();
        }
        
        final Document faultDocument = SourceHelper.documentBuilderThreadLocal.get().newDocument();
        
        final Element faultElement = faultDocument.createElementNS("http://schemas.xmlsoap.org/soap/envelope/", "soapenv:Fault");
        faultDocument.appendChild(faultElement);
        
        final Element faultCodeElement = faultDocument.createElement("faultcode");
        final String faultCodePrefix = faultDocument.lookupPrefix(faultCode.getNamespaceURI());
        if (faultCodePrefix != null) {
            faultCodeElement.appendChild(faultDocument.createTextNode(faultCodePrefix + ":" + faultCode.getLocalPart()));
        }
        else {
            final Attr attr = faultDocument.createAttributeNS("http://www.w3.org/2000/xmlns/", "xmlns:tns");
            attr.appendChild(faultDocument.createTextNode(faultCode.getNamespaceURI()));
            faultCodeElement.setAttributeNode(attr);
            faultCodeElement.appendChild(faultDocument.createTextNode("tns:" + faultCode.getLocalPart()));
        }
        
        final Element faultStringElement = faultDocument.createElement("faultstring");
        faultStringElement.appendChild(faultDocument.createTextNode(faultString));
        
        final Element faultDetailElement = faultDocument.createElement("detail");
        faultDetailElement.appendChild(faultDocument.createTextNode(details));
        
        final Element faultActorElement = faultDocument.createElement("faultactor");
        if (faultActor != null) {
            faultActorElement.appendChild(faultDocument.createTextNode(faultActor.toString()));
        }
        
        faultElement.appendChild(faultCodeElement);
        faultElement.appendChild(faultStringElement);
        faultElement.appendChild(faultActorElement);
        faultElement.appendChild(faultDetailElement);
            
        return faultDocument;
    }

	/**
     * Create a Source from the String. The string must be in UTF-8 format,
     * otherwise, the Source is not created.
     * 
     * @param msg
     *            an UTF-8 String
     * @return the Source built from the String, null if not an UTF-8
     * @deprecated use the method ${@link #createSource(Document)}
     *             instead.
     */
    @Deprecated
    public static Source createSource(final String msg)
			throws PEtALSCFException {
		StreamSource source = new StreamSource();
		try {
			byte[] msgByte = msg.getBytes("UTF-8");
			ByteArrayInputStream in = new ByteArrayInputStream(msgByte);
			source.setInputStream(in);
		} catch (IOException e) {
			throw new PEtALSCFException(e);
		}
		return source;
	}

	/**
     * Create a String from a Source. The String will be encoded with the
     * default locale.
     * 
     * @param s
     * @return
     * @throws PEtALSCFException
     * @deprecated use the method ${@link #createDocument(Source)}
     */
	@Deprecated
    public static String createString(final Source s) throws PEtALSCFException {

		String result = null;
		try {
			if (s instanceof StreamSource
					&& ((StreamSource) s).getInputStream().markSupported()) {
				final StreamSource ss = (StreamSource) s;
				ss.getInputStream().reset();
				byte[] buf = new byte[ss.getInputStream().available()];
				final BufferedInputStream bis = new BufferedInputStream(ss
						.getInputStream());
				bis.read(buf);
				result = new String(buf);
				ss.getInputStream().reset();
			} else {
				final StringWriter buffer = new StringWriter();
				final Result sresult = new StreamResult(buffer);
				SourceHelper.transformerWithoutXmlDeclarationThreadLocal.get().transform(s, sresult);
				result = buffer.toString();
			}
		} catch (IOException e) {
			throw new PEtALSCFException(e);
		} catch (TransformerException e) {
			throw new PEtALSCFException(e);
		}
		return result;
	}

	/**
     * Create a Source from a DOM document.
     * 
     * @param document
     *            A DOM document
     * @return the Source built from the DOM document
     */
    public static Source createSource(final Document document) {
    	document.normalizeDocument();
        return new DOMSource(document);
    }

    /**
     * Create a DOM Document from a Source.
     * 
     * @param s the ${link {@link Source}}
     * @return the DOM document
     * @throws PEtALSCFException
     */
    public static Document createDocument(final Source s) throws PEtALSCFException {
        
        try {
            final DOMResult domResult = new DOMResult();
            SourceHelper.transformerWithoutXmlDeclarationThreadLocal.get().transform(s, domResult);
            
            Node nodeDocument = domResult.getNode().getFirstChild();
            Document document = nodeDocument.getOwnerDocument();
            return document;
        } catch (TransformerException e) {
            throw new PEtALSCFException(e);
        }
    }

	/**
	 * Create a String from a Source.
	 * 
	 * @param s
	 * @return the String in UTF-8 representation without the start and end tags
	 * @throws PEtALSCFException
	 * @deprecated Don't use this methode. Prefer ${@link #createDocument(Source)}
	 */
    @Deprecated
	public static String createStringContent(final Source s)
			throws PEtALSCFException {
		String content = createString(s);
		int index = content.indexOf(START_TAG);
		String result = content.substring(index + START_TAG.length());
		return result.substring(0, result.indexOf(END_TAG));
	}

}
